// TWI Master driver for nRF52840 | gat 11/04/23
#include "TWIM.h"
#include "nrf52840.h"
#include "nrf52840_bitfields.h"
#include <stddef.h> // for NULL

#include "semphr.h"

static NRF_TWIM_Type* GetReg(TWIM_Channel ch)
{
	switch(ch)
	{
		default:
		case TWI0:
		return NRF_TWIM0;
		case TWI1:
		return NRF_TWIM1;
	}
}

SemaphoreHandle_t TWI0_Mutex;
static StaticSemaphore_t MutexBuffer;
static TaskHandle_t WakeMeUpBuffer[]={NULL,NULL};

void TWIM_Setup(TWIM_Config cfg)
{
	NRF_TWIM_Type* reg = GetReg(cfg.Channel);
	reg->ENABLE = 6;
	reg->PSEL.SCL = cfg.SCL_Pin + (cfg.SCL_Port<<5U);
	reg->PSEL.SDA = cfg.SDA_Pin + (cfg.SDA_Port<<5U);
	reg->FREQUENCY = cfg.Freq;
	reg->INTENSET = TWIM_INTENSET_STOPPED_Msk | TWIM_INTENSET_ERROR_Msk;
	TWI0_Mutex = xSemaphoreCreateBinaryStatic(&MutexBuffer); // Binary semaphore and not mutex because we cant give token to mutex in ISR
	(void)xSemaphoreGive(TWI0_Mutex); // Start with a token in the binary semaphore or no one can own TWI0
}

void TWIM_Tx(TWIM_Channel ch, unsigned char addr, unsigned short length,unsigned char* ptr, TaskHandle_t WakeMeUp)
{
	NRF_TWIM_Type* reg = GetReg(ch);
	WakeMeUpBuffer[reg == NRF_TWIM0 ? 0 : 1] = WakeMeUp;
	reg->ENABLE = 6;
	reg->ADDRESS = addr;
	reg->EVENTS_ERROR		=0; 
	reg->EVENTS_STOPPED	=0;
	reg->TXD.LIST 	= 0;
	reg->TXD.PTR    = (uint32_t)ptr;
	reg->TXD.MAXCNT = length;
	reg->SHORTS 		= TWIM_SHORTS_LASTTX_STOP_Msk;
	reg->TASKS_RESUME = 1;
	reg->TASKS_STARTTX = 1;
}

void TWIM_Rx(TWIM_Channel ch, unsigned char addr,unsigned short length,unsigned char* ptr, TaskHandle_t WakeMeUp)
{
	NRF_TWIM_Type* reg = GetReg(ch);
	WakeMeUpBuffer[reg == NRF_TWIM0 ? 0 : 1] = WakeMeUp;
	reg->ENABLE = 6;
	reg->ADDRESS = addr;
	reg->EVENTS_ERROR		=0; 
	reg->EVENTS_STOPPED	=0;
	reg->RXD.LIST 	= 0;
	reg->RXD.PTR    = (uint32_t)ptr;
	reg->RXD.MAXCNT = length;
	reg->SHORTS 		= TWIM_SHORTS_LASTRX_STOP_Msk;
	reg->TASKS_RESUME = 1;
	reg->TASKS_STARTRX = 1;
}

static TWIM_Status Port_Status[2];

void TWIM_GetStatus(TWIM_Channel ch)
{
	TWIM_Status ret;
	NRF_TWIM_Type* reg = GetReg(ch);
	ret.Overrun = reg->ERRORSRC & 1U;
	ret.AddressNACK = reg->ERRORSRC & 2U;
	ret.DataNACK = reg->ERRORSRC & 4U;
	reg->ERRORSRC = 7U;
	Port_Status[ch == TWI0 ? 0 : 1] = ret;
}

TWIM_Status TWIM_ReadStatus(TWIM_Channel ch)
{
	if(ch == TWI0) return Port_Status[0];
	return Port_Status[1];
}

void EndOfTx(unsigned char port) // Wakes up blocking functions
{
	BaseType_t xHigherPriorityTaskWoken = pdFALSE,xHigherPriorityTaskWoken2 = pdFALSE;
	if(WakeMeUpBuffer[port] != NULL) vTaskNotifyGiveFromISR(WakeMeUpBuffer[port], &xHigherPriorityTaskWoken);
	(void)xSemaphoreGiveFromISR(TWI0_Mutex, &xHigherPriorityTaskWoken2); // Free port
	if(xHigherPriorityTaskWoken == pdFALSE) xHigherPriorityTaskWoken = xHigherPriorityTaskWoken2;
	portYIELD_FROM_ISR(xHigherPriorityTaskWoken); // Yield scheduler if needed
}

void TWIM1_IRQ(void)
{
	if(NRF_TWIM1->EVENTS_ERROR)
	{
		NRF_TWIM1->EVENTS_ERROR=0; 
		NRF_TWIM1->TASKS_STOP=1;
		//NRF_TWIM1->ERRORSRC;
	}
	else if (NRF_TWIM1->EVENTS_STOPPED)
	{
		NRF_TWIM1->EVENTS_STOPPED=0;
	}
	//EndOfTx(1);
}

void SPIM0_SPIS0_TWIM0_TWIS0_SPI0_TWI0_IRQHandler(void)
{
	if(NRF_TWIM0->EVENTS_ERROR)
	{
		NRF_TWIM0->EVENTS_ERROR=0; 
		NRF_TWIM0->TASKS_STOP=1;
	}
	else if (NRF_TWIM0->EVENTS_STOPPED)
	{
		NRF_TWIM0->EVENTS_STOPPED=0;	
		TWIM_GetStatus(TWI0);
		EndOfTx(0);
	}
}


