// Differential Pressure Transducer Middleware | gat 12/04/23
#include "DPT.hpp"
#include "TWIM.h"
#include "Timer.h"
#include "GPIOTE.h"
#include "GPIO.h"
#include "HSIS.h"
#include "ports.h"
#include <cstring> //memcpy

#define EOC_ENABLE // Comment to disable EOC
#define TWI_Address 41
#define Timeout_Start (TickType_t) 100U
#define Timeout_DataRead (TickType_t) 100U

enum DPT_TWI_Command{
	Start_Single = 0xAA,
	Start_Average2	= 0xAC,
	Start_Average4	= 0xAD,
	Start_Average8	= 0xAE,
	Start_Average16 = 0xAF
};

// Stored outside the class because its C and C++ shared static variables
extern SemaphoreHandle_t TWI0_Mutex;
static TaskHandle_t* parent_query_task = NULL;

static unsigned char data_buffer[2][Data_size];
static unsigned char data_buffer_read_idx;
static unsigned char data_buffer_write_idx;

void DPT::Setup(TaskHandle_t* query_task)
{
	parent_query_task = query_task;
	TWIM_Config I2C_Config = {
		.Channel = DPT_TWIM_Channel,
		.Freq = K250,
		.SDA_Pin = I2C_SDA_Pin,
		.SDA_Port = I2C_SDA_Port,
		.SCL_Pin = I2C_SCL_Pin,
		.SCL_Port = I2C_SCL_Port
	};
	TWIM_Setup(I2C_Config);
	
	#ifdef EOC_ENABLE
	GPIOTE_Config GPIOTE_Config = {
		.Mode = GPIOTE_Event,
		.Polarity = GPIOTE_LowToHigh,
		.Channel = GPIOTE_Ch1,
		.Pin = DPT_EOC_Pin,
		.Port = DPT_EOC_Port,
		.Event_Callback = EndOfConversion
	};
	GPIOTE_Setup(GPIOTE_Config);
	#endif
}

DPT_Return DPT::ReadData()
{
	if(xSemaphoreTake(TWI0_Mutex, 1U) == pdFALSE) return DPT_MutexError;
	TWIM_Rx(DPT_TWIM_Channel, TWI_Address, Data_size, GetWriteDataPtr(), NULL);
	SwapDataPtr();
	return DPT_Success;
}

DPT_Return DPT::ReadDataBlocking()
{
	DPT_Return ret;
	if(xSemaphoreTake(TWI0_Mutex, 1U) == pdFALSE) return DPT_MutexError;
	TWIM_Rx(DPT_TWIM_Channel, TWI_Address, Data_size, GetWriteDataPtr(), xTaskGetCurrentTaskHandle());
	ret = ulTaskNotifyTake(pdTRUE, Timeout_DataRead) ? DPT_Success : DPT_Timeout;
	SwapDataPtr();
	return ret;
}

DPT_Return DPT::Start()
{
	unsigned char command = Start_Single;
	if(xSemaphoreTake(TWI0_Mutex, portMAX_DELAY) == pdFALSE) return DPT_MutexError;
	TWIM_Tx(DPT_TWIM_Channel, TWI_Address, 1U, &command, NULL);
	return DPT_Success;
}

DPT_Return DPT::StartBlocking()
{
	DPT_Return ret;
	unsigned char command = Start_Single;
	if(xSemaphoreTake(TWI0_Mutex, 1U) == pdFALSE) return DPT_MutexError;
	TWIM_Tx(DPT_TWIM_Channel, TWI_Address, 1U, &command, xTaskGetCurrentTaskHandle());
	ret = ulTaskNotifyTake(pdTRUE, Timeout_Start) ? DPT_Success : DPT_Timeout;
	return ret;
}

void DPT::SwapDataPtr()
{
	unsigned char tmp = data_buffer_read_idx;
	data_buffer_read_idx = data_buffer_write_idx;
	data_buffer_write_idx = tmp;
}

unsigned char* DPT::GetReadDataPtr() { return data_buffer[data_buffer_read_idx]; }

unsigned char* DPT::GetWriteDataPtr() { return data_buffer[data_buffer_write_idx]; }

//#define TEST1
//#define ENABLE_NORMALISATION

#ifdef ENABLE_NORMALISATION
#define NORMALISATION_COEF 3.205f // Task #12673
#define NORMALISATION_OFFSET 0x800000
typedef union 
{
	unsigned char bulk[4U];
	int value;
} DPT_Normalized;
#endif // ENABLE_NORMALISATION


void DPT::GetSample(unsigned char* data_ptr)
{
	#ifdef TEST1
	unsigned char dummyTEST[] = {0x12,0x34,0x56};
	memcpy(data_ptr, dummyTEST, sizeof(dummyTEST));
	#else
	unsigned char* dat = GetReadDataPtr();
	#ifndef ENABLE_NORMALISATION
	memcpy(data_ptr, dat + 1, 3);
	#else
	DPT_Normalized norm;
	norm.bulk[0] = dat[3];
	norm.bulk[1] = dat[2];
	norm.bulk[2] = dat[1];
	norm.value -= NORMALISATION_OFFSET;
	norm.value /= NORMALISATION_COEF;
	norm.value += NORMALISATION_OFFSET;
	dat[0] = norm.bulk[2];
	dat[1] = norm.bulk[1];
	dat[2] = norm.bulk[0];
	#endif // ENABLE NORMALISATION
	#endif // TEST
}

void DPT::GetStatus()
{
	/*
	unsigned char status;
	TWIM_Rx(DPT_TWIM_Channel, TWI_Address, 1U, &status, );
	(void)status;
	*/
}

void DPT::EndOfConversion()
{
	BaseType_t xHigherPriorityTaskWoken = pdFALSE;
	io_tog(DBG_MCU_RX_Port, DBG_MCU_RX_Pin);
	if(parent_query_task != NULL)
	{
		vTaskNotifyGiveFromISR(*parent_query_task, &xHigherPriorityTaskWoken);
		portYIELD_FROM_ISR(xHigherPriorityTaskWoken);
	}
}
