#include "easyDAC.hpp"
#include "Acquisition.hpp"
#include "FreeRTOS.h"
#include "task.h"
#include "ports.h"
#include "SPIS.h"
#include "HSIS.h"
#include "NVIC.h"
#include "GPIO.h"
#include "bootload.h"
#include "FlashStorage.hpp"
#include "Task_Priority.h"
#include "neb.hpp"

typedef enum : unsigned char
{
	Processing_Action,
	Success,
	Failure,
	Need_Aux_Supply
} Action_ack;

typedef enum : unsigned char
{
	Entry_State,
	Read_State = 0x01,
	Read_Description = 0x03,
	POD_Config = 0x05,
	Implant_Config = 0x07,
	Start_Acquisition = 0x10,
	Stop_Acquisition = 0x11,
	Stop_Start_Neb = 0x13,
	Serial_Number = 0x14,
	Bootload = 0x15,
	Read_Slow_Sig_Cal = 0x16,
	Write_Slow_Sig_Cal = 0x17,
	Set_Neb = 0x18,
	Get_Neb = 0x19,
	Error
}ezDAC_Parsing_State;

#define STACK_SIZE 100
static TaskHandle_t Task;
static StaticTask_t xTaskBuffer;
static StackType_t xStack[STACK_SIZE];


#define POD_type_Mouse 5
#define POD_type_Rat 6
#define Current_POD_type POD_type_Mouse

#define Bulk_Size 256U
static unsigned char Bulk[Bulk_Size];
static unsigned int Bulk_Write_idx;
static unsigned int Bulk_Read_idx;
static void * local_tx_ptr;
static unsigned int local_tx_size;

static ezDAC_Parsing_State parsing_state = Entry_State;
static unsigned char POD_cfg_bulk[16U];
static unsigned char Implant_cfg_bulk[21U];
static unsigned char State_bulk[2U] = 
{
	0x0F,0x00
};
static unsigned char Description_bulk[]=
{
	Current_POD_type, // POD type
	0x11U, 0x22U, 0x33U, 0x44U, 0x55U, 0x66U, 0x77U, 0x88U, // S/N
	0x01U, 0x00U, 0x00U, // CW version
	0x01U, 0x00U, 0x00U // BTL version
};
static unsigned char Detection_bulk[]=
{
	Current_POD_type, // POD type
	0,0,0
};

static unsigned char Neb_bulk[3U];

bool easyDAC::ImplantConfigReceived = false;

extern vF_Acq_Frame Acquisition_Buffer;
extern bool IsAcqEnabled;

static void Reset_Local_Tx(void)
{
	if(IsAcqEnabled)
	{
		local_tx_ptr = Acquisition_Buffer.bulk;
		local_tx_size = 2U;
	}
	else
	{
		local_tx_ptr = Detection_bulk;
		local_tx_size = sizeof(Detection_bulk);
	}
}

void easyDAC::Setup()
{
	SPIS_Config spis_cfg = {
		.Channel = easyDAC_SPI,
		.CPHA = SPIS_Trailing,
		.CPOL = SPIS_ActiveHigh,
		.Endian = SPIS_MSB_First,
		.MOSI_Pin = SPI_MOSI_Pin,
		.MOSI_Port = SPI_MOSI_Port,
		.MISO_Pin = SPI_MISO_Pin,
		.MISO_Port = SPI_MISO_Port,
		.CLK_Pin = SPI_SCLK_Pin,
		.CLK_Port = SPI_SCLK_Port,
		.CS_Pin = SPI_CS_Pin,
		.CS_Port = SPI_CS_Port };
	SPIS_Setup(spis_cfg);
	// Load daemon firmware version from regrets
	int i = 1;
	FlashStorage::Load_Serial_Number(Description_bulk + i);
	i += 8;
	Description_bulk[i++] = FIRMWARE_VERSION_1;
	Description_bulk[i++] = FIRMWARE_VERSION_2;
	Description_bulk[i++] = FIRMWARE_VERSION_3;
	FlashStorage::Load_Daemon_Version(Description_bulk + i, Description_bulk + i + 1, Description_bulk + i + 2);
	Task = xTaskCreateStatic(Reception_Task, "RX", STACK_SIZE, NULL, RX_PRIORITY, xStack, &xTaskBuffer);
}

void easyDAC::Reception_Task(void* context)
{
	unsigned int rxed_count = 0;
	(void)context;
	SPIS_Rx_Callback_Register(easyDAC_SPI, Port_IRQ);
	Reset_Local_Tx();
	while(1)
	{
		SPIS_Update(easyDAC_SPI, local_tx_ptr, local_tx_size, Bulk, Bulk_Size);
		Bulk_Read_idx = Bulk_Write_idx = 0;
		(void)ulTaskNotifyTake(pdTRUE, portMAX_DELAY); // Woken up by SPI rx end
		rxed_count = SPIS_Get_Rxed_Count(easyDAC_SPI); // How many bytes received
		Bulk_Write_idx += rxed_count; // Move cursor
		while(Bulk_Read_idx < Bulk_Write_idx) // Treat all rxed bytes one by one
			Parse_Byte(Bulk[Bulk_Read_idx++]);
	}
}

void easyDAC::Parse_Byte(unsigned char b)
{
	static unsigned char byte_count = 0;
	switch(parsing_state) // RX switch
	{
	case Entry_State:
		parsing_state = (ezDAC_Parsing_State)b;
		byte_count = 0;
		break;
	case POD_Config:
		POD_cfg_bulk[byte_count++] = b;
		if(byte_count == sizeof(POD_cfg_bulk)) parsing_state = Entry_State;
		break;
	case Implant_Config:
		Implant_cfg_bulk[byte_count++] = b;
		if(byte_count == sizeof(Implant_cfg_bulk))
		{
			ImplantConfigReceived = true;
			parsing_state = Entry_State;
		}
		break;
	case Serial_Number:
		Description_bulk[1 + (byte_count++)] = b; // Update S/N
		if(byte_count == 8U) {
			parsing_state = Entry_State;
			FlashStorage::Store_Serial_Number(Description_bulk + 1);
		}
		break;
	case Bootload:
		go_to_bootloader(b);
		break;
	case Stop_Start_Neb:
		// TODO
		byte_count ++;
	 	if(byte_count == 8U) parsing_state = Entry_State;
		break;
	case Set_Neb:
		Neb_bulk[byte_count ++] = b;
	 	if(byte_count == sizeof(Neb_bulk)) 
		{
			Acquisition_Buffer.Action_status = Processing_Action;
			parsing_state = Entry_State;
			NebInfo cfg = { 
				.frequency = Neb_bulk[0] + (Neb_bulk[1] << 8),
				.duty_cycle = Neb_bulk[2] };
			if(!Neb::Update(cfg)) // Neb off
			{
				Acquisition_Buffer.Action_status = Success;
			}
			else // Neb on
			{
				if(Neb::Detect_Supply())
				{
					Acquisition_Buffer.Action_status = Success;
					Neb::Enable();
				}
				else
				{
					Acquisition_Buffer.Action_status = Need_Aux_Supply;
					Neb::Disable();
				}
			}
		}
		break;
	default:
	case Error:
		parsing_state = Entry_State;
		break;
	}
	Reset_Local_Tx();
	switch (parsing_state) // TX switch
	{
	case Read_State:
		parsing_state = Entry_State;
		Acquisition::Detect(State_bulk);
		local_tx_ptr = State_bulk;
		local_tx_size = sizeof(State_bulk);
		break;
	case Read_Description:
		parsing_state = Entry_State;
		local_tx_ptr = Description_bulk;
		local_tx_size = sizeof(Description_bulk);
		break;
	case Start_Acquisition:
		parsing_state = Entry_State;
		Acquisition::Enable();
		break;
	case Stop_Acquisition:
		parsing_state = Entry_State;
		Acquisition::Disable();
		break;
	case Get_Neb:
	{
		parsing_state = Entry_State;
		NebInfo info = Neb::GetInfo();
		Neb_bulk[0] = (unsigned char)(info.frequency & 0xFF);
		Neb_bulk[1] = (unsigned char)((info.frequency & 0xFF)>>8U);
		Neb_bulk[2] = (unsigned char)(info.duty_cycle & 0xFF);
		local_tx_ptr = Neb_bulk;
		local_tx_size = sizeof(Neb_bulk);
	}
		break;
	default:
		break;
	}
}

unsigned char* easyDAC::GetSignalsConfig() { return POD_cfg_bulk; }

unsigned char* easyDAC::GetImplantConfig() { return Implant_cfg_bulk; }

void easyDAC::Prepare_Acq_Frame(unsigned char size) {	SPIS_Update(easyDAC_SPI, Acquisition_Buffer.bulk, size, Bulk, Bulk_Size);	}

void easyDAC::Port_IRQ()
{
	BaseType_t xHigherPriorityTaskWoken = pdFALSE;
	io_set(TRIG_Port, TRIG_Pin); // TRIG = 1 w/e at this point
	vTaskNotifyGiveFromISR(Task, &xHigherPriorityTaskWoken);
	portYIELD_FROM_ISR(xHigherPriorityTaskWoken);
}
